import nbformat
import pytest
from nbformat import NotebookNode

SOURCE = """\
# #fig
import matplotlib.pyplot as plt
plt.plot([1, 10, 100], [-1, 0, 1])
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [nbformat.v4.new_code_cell(SOURCE)]
    execute(nb)
    return nb


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "fig")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["image/png"].startswith("iVBO")


def test_mime_content(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    data = get_mime_content(nb, "fig")
    assert isinstance(data, tuple)
    assert len(data) == 2
    assert data[0] == "image/png"
    assert isinstance(data[1], bytes)



import nbformat
import pytest
from nbformat import NotebookNode


@pytest.mark.parametrize(
    ("text", "expected"),
    [
        ("", []),
        ("'", ["'"]),
        ("''", ["''"]),
        ('"', ['"']),
        ('""', ['""']),
        (" ", []),
        ("   ", []),
        ("=", ["="]),
        (" =", ["="]),
        ("= ", ["="]),
        ("abc", ["abc"]),
        ("αβ γδ", ["αβ", "γδ"]),
        (" a  b  c ", ["a", "b", "c"]),
        ('"a b c"', ['"a b c"']),
        ("'a b c'", ["'a b c'"]),
        ("`a b c`", ["`a b c`"]),
        ("a 'b c' d", ["a", "'b c'", "d"]),
        ("a `b c` d", ["a", "`b c`", "d"]),
        ('a "b c" d', ["a", '"b c"', "d"]),
        (r"a 'b \'c\' d' e", ["a", r"'b \'c\' d'", "e"]),
        ("a=b", ["a=b"]),
        ("a = b", ["a=b"]),
        ("a = b c = d", ["a=b", "c=d"]),
        ("a = b c =", ["a=b", "c", "="]),
        ("a='b c' d = 'e f'", ["a='b c'", "d='e f'"]),
    ],
)
def test_split(text, expected):
    from nbstore.markdown import split

    assert list(split(text)) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("", ""),
        ("a", '"a"'),
        ("a b", '"a b"'),
        ("a b c", '"a b c"'),
    ],
)
def test_quote(value, expected):
    from nbstore.markdown import _quote

    assert _quote(value) == expected


def test_quote_single():
    from nbstore.markdown import _quote

    assert _quote('a "b" c') == "'a {} c'".format('"b"')


SOURCE = """\
![a](b.ipynb){ #c .s k=v}

abc `![a](b){c}`

```python
![a](b){c}
```

``` {.text #id a = 'b c'}
xyz
```

```nobody
```

```
noattr
```
"""


@pytest.fixture(scope="module")
def elements():
    from nbstore.markdown import parse

    return list(parse(SOURCE))


def test_elements_image(elements):
    from nbstore.markdown import Image

    x = elements[0]
    assert isinstance(x, Image)
    assert x.alt == "a"
    assert x.url == "b.ipynb"
    assert x.identifier == "c"
    assert x.classes == [".s"]
    assert x.attributes == {"k": "v"}


def test_elements_code_block(elements):
    from nbstore.markdown import CodeBlock

    x = elements[2]
    assert isinstance(x, CodeBlock)
    assert x.source == "![a](b){c}"
    assert x.identifier == ""
    assert x.classes == ["python"]
    assert x.attributes == {}


def test_elements_code_block_with_attributes(elements):
    from nbstore.markdown import CodeBlock

    x = elements[4]
    assert isinstance(x, CodeBlock)
    assert x.source == "xyz"
    assert x.identifier == "id"
    assert x.classes == [".text"]
    assert x.attributes == {"a": "b c"}


def test_elements_code_block_without_body(elements):
    from nbstore.markdown import CodeBlock

    x = elements[6]
    assert isinstance(x, CodeBlock)
    assert x.source == ""
    assert x.identifier == ""
    assert x.classes == ["nobody"]
    assert x.attributes == {}


def test_elements_code_block_without_attributes(elements):
    from nbstore.markdown import CodeBlock

    x = elements[8]
    assert isinstance(x, CodeBlock)
    assert x.source == "noattr"
    assert x.identifier == ""
    assert x.classes == []
    assert x.attributes == {}


@pytest.mark.parametrize(
    ("index", "expected"),
    [(1, "\n\nabc `![a](b){c}`\n\n"), (3, "\n\n"), (5, "\n\n")],
)
def test_elements_str(elements, index, expected):
    x = elements[index]
    assert isinstance(x, str)
    assert x == expected


def test_join(elements):
    x = [x if isinstance(x, str) else x.text for x in elements]
    assert "".join(x) == SOURCE


def test_iter_parts():
    from nbstore.markdown import Element

    x = Element("", "id", ["a", "b"], {"k": "v"})
    assert list(x.iter_parts()) == ["a", "b", 'k="v"']
    assert list(x.iter_parts(include_identifier=True)) == ["#id", "a", "b", 'k="v"']


def test_iter_parts_exclude_attributes():
    from nbstore.markdown import Element

    x = Element("", "id", ["a", "b"], {"k": "v", "l": "w"})
    assert list(x.iter_parts(exclude_attributes=["l"])) == ["a", "b", 'k="v"']


@pytest.mark.parametrize(
    ("markdown", "expected"),
    [
        ("```python {a #id1 b}\nprint(1)\n```", (["python", "a", "b"], "id1")),
        ("```{.python #id2 b}\nprint(2)\n```", ([".python", "b"], "id2")),
        ("```{python #id3 {a} }\nprint(3)\n```", (["python", "{a}"], "id3")),
        ("```bash {#id4}\necho hello\n```", (["bash"], "id4")),
        ("```python\nprint(4)\n```", (["python"], "")),
        ('```{python #id3 "{a}" }\np\n```', (["python", '"{a}"'], "id3")),
    ],
)
def test_markdown_code_blocks(markdown, expected):
    from nbstore.markdown import CodeBlock, parse

    x = next(parse(markdown))
    assert isinstance(x, CodeBlock)
    assert x.classes == expected[0]
    assert x.identifier == expected[1]


def test_code_block_url():
    from nbstore.markdown import CodeBlock, parse

    x = next(parse("```python a.ipynb#id c\nprint(1)\n```\n"))
    assert isinstance(x, CodeBlock)
    assert x.source == "print(1)"
    assert x.url == "a.ipynb"
    assert x.identifier == "id"
    assert x.classes == ["python", "c"]


def test_image_code():
    from nbstore.markdown import Image, parse

    x = next(parse("![alt](a.ipynb){#id `co de` b}"))
    assert isinstance(x, Image)
    assert x.source == "co de"
    assert x.url == "a.ipynb"
    assert x.identifier == "id"
    assert x.classes == ["b"]


def test_image_code_eq():
    from nbstore.markdown import Image, parse

    x = next(parse("![alt](a.ipynb){#id `a=1;b=2` b}"))
    assert isinstance(x, Image)
    assert x.source == "a=1;b=2"
    assert x.url == "a.ipynb"
    assert x.identifier == "id"
    assert x.classes == ["b"]


SOURCE_LANG = """\

```python #_
import matplotlib.pyplot as plt

def plot(x):
    plt.plot([x])
```

```julia #_
println("hello")
```

```python #plot-1
plot(1)
```

![alt](.md){#plot-1}

"""


def test_get_language():
    from nbstore.markdown import get_language

    assert get_language(SOURCE_LANG) == "python"


def test_get_language_none():
    from nbstore.markdown import get_language

    assert get_language("") == ""


SOURCE_LANG_AFTER = """\

![alt](.md){#plot-1}

```python #_
println("hello")
```

```julia #plot-1
plot(1)
```
"""


def test_get_language_after():
    from nbstore.markdown import get_language

    assert get_language(SOURCE_LANG_AFTER) == "julia"


@pytest.mark.parametrize(
    ("language", "expected"),
    [("julia", True), ("python", False), (".julia", True), (".python", False)],
)
def test_is_target_code_block(language, expected):
    from nbstore.markdown import CodeBlock, is_target_code_block

    code = CodeBlock("", "id", [language], {})
    assert is_target_code_block(code, "julia") == expected


def test_is_target_code_block_no_language():
    from nbstore.markdown import is_target_code_block

    assert not is_target_code_block("", "")


def test_is_target_code_block_str():
    from nbstore.markdown import is_target_code_block

    assert not is_target_code_block("python", "python")


def test_is_target_code_block_no_identifier():
    from nbstore.markdown import CodeBlock, is_target_code_block

    code = CodeBlock("", "", [], {})
    assert not is_target_code_block(code, "python")


SOURCE_NOTEBOOK = """\

```python #_
import matplotlib.pyplot as plt

def plot(x):
    plt.plot([x])
```

```julia #_
println("hello")
```

```python #plot-1
plot(1)
```

```{.python #plot-2}
plot(2)
```

```python
plot(3)
```

![alt](.md){#plot-2}

"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.markdown import new_notebook

    return new_notebook(SOURCE_NOTEBOOK)


@pytest.fixture(scope="module")
def test_len(nb: NotebookNode):
    assert len(nb["cells"]) == 4


def test_source_1(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-1") == "plot(1)"


def test_source_2(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-2") == "plot(2)"


def test_language_default():
    from nbstore.notebook import get_language

    nb = nbformat.v4.new_notebook()
    assert get_language(nb, "julia") == "julia"


def test_language_error():
    from nbstore.markdown import new_notebook

    with pytest.raises(ValueError, match="language not found"):
        new_notebook("hello")


SOURCE_INDENT_IMAGE = """\
![alt](.md){#id-0}
 ![alt](.md){#id-1}

def fgi
    ![alt](.md){#id-4}
jkl
a ![alt](.md){#id-a} ![alt](.md){#id-b}
"""


def test_indent_image():
    from nbstore.markdown import Image, parse

    elems = [e for e in parse(SOURCE_INDENT_IMAGE) if isinstance(e, Image)]
    assert len(elems) == 5
    assert elems[0].indent == ""
    assert elems[1].indent == " "
    assert elems[1].text == " ![alt](.md){#id-1}"
    assert elems[2].indent == "    "
    assert elems[2].text == "    ![alt](.md){#id-4}"
    assert elems[3].indent == ""
    assert elems[3].text == "![alt](.md){#id-a}"
    assert elems[4].indent == ""
    assert elems[4].text == "![alt](.md){#id-b}"


def test_join_image():
    from nbstore.markdown import parse

    elems = parse(SOURCE_INDENT_IMAGE)
    x = [e if isinstance(e, str) else e.text for e in elems]
    assert "".join(x) == SOURCE_INDENT_IMAGE


SOURCE_INDENT_CODE_BLOCK = """\
```python
a
  b
```
a

    ```python
    a
      b
    ```
    d
"""


def test_indent_code_block():
    from nbstore.markdown import CodeBlock, parse

    elems = [e for e in parse(SOURCE_INDENT_CODE_BLOCK) if isinstance(e, CodeBlock)]
    assert len(elems) == 2
    assert elems[0].indent == ""
    assert elems[0].source == "a\n  b"
    assert elems[0].text.startswith("```python")
    assert elems[1].indent == "    "
    assert elems[1].source == "a\n  b"
    assert elems[1].text.startswith("    ```python")


def test_join_code_block():
    from nbstore.markdown import parse

    elems = parse(SOURCE_INDENT_CODE_BLOCK)
    x = [e if isinstance(e, str) else e.text for e in elems]
    assert "".join(x) == SOURCE_INDENT_CODE_BLOCK


def test_comment():
    from nbstore.markdown import parse

    text = "<!-- ![alt](a.py){#id} -->"
    assert next(parse(text)) == text



"""Extract and manipulate content from Jupyter notebooks.

nbstore provides tools for reading, processing, and storing notebook content,
with support for multiple formats including .ipynb, .py, and .md files.
It facilitates extraction of specific cells, data conversion, and
formatting of visualization outputs.

Main components:
- Store: Main interface for reading and managing notebook files
- Formatter: Controls output format for visualization libraries
- Content parsing for Python and Markdown files
"""

from .formatter import set_formatter
from .store import Store, read

__all__ = ["Store", "read", "set_formatter"]



import pytest
from nbformat import NotebookNode

SOURCE = """\

def plot(x: int):
    print(x)

# %% #plot-1
plot(1)

if __name__ == "__main__":
    # %% #plot-2

    plot(2)

if __name__ == "__main__":
    #
    # %% #plot-3

    plot(3)

# %% #plot-4

plot(4)

"""


@pytest.fixture
def text():
    return SOURCE


@pytest.fixture
def blocks(text):
    from nbstore.python import _iter_main_blocks

    return list(_iter_main_blocks(text))


def test_blocks_0(blocks: list[str]):
    assert blocks[0].startswith("\ndef plot")
    assert blocks[0].endswith("plot(1)\n")


def test_blocks_1(blocks: list[str]):
    assert blocks[1].startswith("# %% #plot-2")
    assert blocks[1].endswith("plot(2)\n")


def test_blocks_2(blocks: list[str]):
    assert blocks[2].startswith("#\n# %% #plot-3")
    assert blocks[2].endswith("plot(3)\n")


def test_blocks_3(blocks: list[str]):
    assert blocks[3].startswith("# %% #plot-4")
    assert blocks[3].endswith("plot(4)\n\n")


@pytest.fixture
def sources(text):
    from nbstore.python import parse

    return list(parse(text))


def test_sources(sources: list[str]):
    assert len(sources) == 6


def test_sources_0(sources: list[str]):
    assert sources[0] == "\ndef plot(x: int):\n    print(x)"


def test_sources_1(sources: list[str]):
    assert sources[1] == "# %% #plot-1\nplot(1)"


def test_sources_2(sources: list[str]):
    assert sources[2] == "# %% #plot-2\n\nplot(2)"


def test_sources_3(sources: list[str]):
    assert sources[3] == "#"


def test_sources_4(sources: list[str]):
    assert sources[4] == "# %% #plot-3\n\nplot(3)"


def test_sources_5(sources: list[str]):
    assert sources[5] == "# %% #plot-4\n\nplot(4)"


SOURCE_NOTEBOOK = """\
def plot(x: int):
    print(x)  # noqa: T201


# %% #plot-1
plot(1)

if __name__ == "__main__":
    # %% #plot-2

    plot(2)

# %% #plot-3
plot(3)

if __name__ == "__main__":
    #
    # %% #plot-4

    plot(4)

# %% #plot-5

plot(5)
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.python import new_notebook

    return new_notebook(SOURCE_NOTEBOOK)


def test_len(nb: NotebookNode):
    assert len(nb["cells"]) == 7


def test_cell_0(nb: NotebookNode):
    assert nb.cells[0]["source"].startswith("def plot(x: int):\n    print(x)")
    assert nb.cells[4]["source"] == "#"


def test_cell_1(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-1") == "plot(1)"


def test_cell_2(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-2") == "\nplot(2)"


def test_cell_3(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-3") == "plot(3)"


def test_cell_4(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-4") == "\nplot(4)"


def test_cell_5(nb: NotebookNode):
    from nbstore.notebook import get_source

    assert get_source(nb, "plot-5") == "\nplot(5)"



"""Convert Markdown files to Jupyter notebooks.

This module provides utilities for parsing Markdown files and converting them
to notebook format, with support for code blocks, images, and custom attributes.
It implements a robust parser for Markdown syntax including code blocks with
language specification and attributes.
"""

from __future__ import annotations

import re
import textwrap
from dataclasses import dataclass
from functools import cache
from itertools import takewhile
from typing import TYPE_CHECKING, ClassVar, TypeGuard

import nbformat

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Self

    from nbformat import NotebookNode


def _split(text: str) -> Iterator[str]:
    """Split text into parts while respecting quoted strings.

    Splits text by spaces, treating quoted strings as single parts, and
    handling escaped quotes correctly.

    Args:
        text (str): The text to split.

    Yields:
        str: The split parts.
    """
    in_quotes = {'"': False, "'": False, "`": False}

    chars = list(text)
    start = 0

    for cursor, char in enumerate(chars):
        if cursor > 0 and chars[cursor - 1] == "\\":
            continue

        for q, in_ in in_quotes.items():
            if char == q:
                if in_:
                    yield text[start : cursor + 1]
                    start = cursor + 1
                in_quotes[q] = not in_

        if char == " ":
            if not any(in_quotes.values()):
                if start < cursor:
                    yield text[start:cursor]
                start = cursor + 1

    if start < len(text):
        yield text[start:]


def split(text: str) -> Iterator[str]:
    """Split text into parts, handling key=value pairs.

    Builds on _split by additionally recognizing and combining key=value
    pairs into single parts.

    Args:
        text (str): The text to split.

    Yields:
        str: The split parts.
    """
    parts = list(_split(text))

    start = 0
    for cursor, part in enumerate(parts):
        if part == "=" and 0 < cursor < len(parts) - 1:
            if start < cursor - 1:
                yield from parts[start : cursor - 1]
            yield f"{parts[cursor - 1]}={parts[cursor + 1]}"
            start = cursor + 2

    if start < len(parts):
        yield from parts[start:]


def _iter(
    pattern: re.Pattern,
    text: str,
    pos: int = 0,
    endpos: int | None = None,
) -> Iterator[re.Match[str] | tuple[int, int]]:
    r"""Iterate over matches of a regex pattern in the given text.

    Search for all occurrences of the specified regex pattern
    in the provided text. Yield the segments of text between matches
    as well as the matches themselves. This allows for processing
    both the matched content and the surrounding text in a single iteration.

    Args:
        pattern (re.Pattern): The compiled regex pattern to search for in the text.
        text (str): The text to search for matches.
        pos (int): The starting position in the text to search for matches.
        endpos (int | None): The ending position in the text to search for matches.

    Yields:
        re.Match | tuple[int, int]: Segments of text and match objects. The segments
        are the parts of the text that are not matched by the pattern, and the
        matches are the regex match objects.

    Examples:
        >>> import re
        >>> pattern = re.compile(r'\d+')
        >>> text = "There are 2 apples and 3 oranges."
        >>> matches = list(_iter(pattern, text))
        >>> matches[0]
        (0, 10)
        >>> matches[1]
        <re.Match object; span=(10, 11), match='2'>
        >>> matches[2]
        (11, 23)
        >>> matches[3]
        <re.Match object; span=(23, 24), match='3'>
        >>> matches[4]
        (24, 33)

    """
    if endpos is None:
        endpos = len(text)

    cursor = pos

    for match in pattern.finditer(text, pos, endpos=endpos):
        start, end = match.start(), match.end()

        if cursor < start:
            yield cursor, start

        yield match

        cursor = end

    if cursor < endpos:
        yield cursor, endpos


def _strip_quotes(value: str) -> str:
    if (value.startswith('"') and value.endswith('"')) or (
        value.startswith("'") and value.endswith("'")
    ):
        return value[1:-1]
    return value


def _quote(value: str) -> str:
    if not value:
        return value
    if '"' in value:
        return f"'{value}'"
    return f'"{value}"'


def _parse(text: str) -> tuple[str, list[str], dict[str, str]]:
    """Parse attribute text into identifier, classes, and attributes.

    Parses a string like "#id .class1 .class2 key1=value1 key2=value2"
    into its components.

    Args:
        text (str): The attribute text to parse.

    Returns:
        tuple[str, list[str], dict[str, str]]: A tuple containing:
            - identifier: The ID (without the '#')
            - classes: List of class names (with or without the '.')
            - attributes: Dictionary of attribute key-value pairs
    """
    identifier = ""
    classes = []
    attributes = {}

    for part in split(text):
        if part.startswith("#"):
            identifier = part[1:]
        elif part.startswith("`") and part.endswith("`"):  # for source syntax
            classes.append(part)
        elif "=" in part:
            key, value = part.split("=", 1)
            attributes[key] = _strip_quotes(value)
        else:
            classes.append(part)  # Do not remove the optional leading dot

    return identifier, classes, attributes


@dataclass
class Matcher:
    pattern: ClassVar[re.Pattern]

    @classmethod
    def from_match(cls, match: re.Match[str]) -> Self | str:
        """Create an element from a regex match.

        Args:
            match (re.Match[str]): The regex match object.

        Returns:
            Self: The created element.

        """
        return match.group(0)

    @classmethod
    def iter_elements(
        cls,
        text: str,
        pos: int = 0,
        endpos: int | None = None,
    ) -> Iterator[Self | str | tuple[int, int]]:
        """Iterate through elements in the text.

        Finds all occurrences of the element pattern in the text and
        yields either element instances or text spans.

        Args:
            text (str): The text to search.
            pos (int): The starting position.
            endpos (int | None): The ending position.

        Yields:
            Self | tuple[int, int]: Element instances or text spans.
        """
        for match in _iter(cls.pattern, text, pos, endpos):
            if isinstance(match, re.Match):
                yield cls.from_match(match)

            else:
                yield match


@dataclass
class Comment(Matcher):
    """A comment in Markdown."""

    pattern: ClassVar[re.Pattern] = re.compile(
        r"<!--(?P<body>.*?)-->",
        re.MULTILINE | re.DOTALL,
    )


@dataclass
class Element(Matcher):
    """Base class for Markdown elements with attributes.

    Represents a structured element in a Markdown document, with support
    for identifiers, classes, attributes, and other properties.

    Attributes:
        pattern: Regular expression pattern to match the element in text.
        text: The original text of the element.
        identifier: The ID of the element.
        classes: List of class names for the element.
        attributes: Dictionary of attribute key-value pairs.
        code: The code content of the element (for code blocks).
        url: The URL of the element (for links and images).
    """

    text: str
    identifier: str
    classes: list[str]
    attributes: dict[str, str]
    source: str = ""
    url: str = ""
    indent: str = ""

    def iter_parts(
        self,
        *,
        include_identifier: bool = False,
        include_classes: bool = True,
        include_attributes: bool = True,
        exclude_attributes: Iterable[str] = (),
    ) -> Iterator[str]:
        """Iterate through the parts of the element's attributes.

        Args:
            include_identifier (bool): Whether to include the identifier.
            include_classes (bool): Whether to include the classes.
            include_attributes (bool): Whether to include the attributes.
            exclude_attributes (Iterable[str]): Attributes to exclude.

        Yields:
            str: The parts of the element's attributes.
        """
        if include_identifier and self.identifier:
            yield f"#{self.identifier}"

        if include_classes:
            yield from self.classes

        if include_attributes:
            for k, v in self.attributes.items():
                if k not in exclude_attributes:
                    yield f"{k}={_quote(v)}"


@dataclass
class CodeBlock(Element):
    """A code block in Markdown.

    Represents a fenced code block, with support for language specification
    and attributes.

    Example:
        ```python #id
        print("Hello, world!")
        ```
    """

    pattern: ClassVar[re.Pattern] = re.compile(
        r"^(?P<pre> *[~`]{3,})(?P<body>.*?)\n(?P=pre)",
        re.MULTILINE | re.DOTALL,
    )

    @classmethod
    def from_match(cls, match: re.Match[str]) -> Self:
        text = match.group(0)
        body = match.group("body")
        pre = match.group("pre")
        indent = "".join(takewhile(str.isspace, pre))

        if "\n" in body:
            attr, source = body.split("\n", 1)
            source = textwrap.dedent(source)
        else:
            attr, source = body, ""

        attr = " ".join(_remove_braces(attr.strip()))
        identifier, classes, attributes = _parse(attr)

        url = ""

        if not identifier:
            for k, cls_ in enumerate(classes):
                if "#" in cls_:
                    url, identifier = cls_.split("#", 1)
                    classes = classes[:k] + classes[k + 1 :]
                    break

        return cls(
            text,
            identifier,
            classes,
            attributes,
            source=source,
            url=url,
            indent=indent,
        )


def _remove_braces(text: str) -> Iterator[str]:
    in_brace = False

    for part in _split(text):
        if part.startswith("{") and part.endswith("}") and in_brace:
            yield part
        elif part.startswith("{") and not in_brace:
            if part.endswith("}"):
                yield part[1:-1]
            else:
                yield part[1:]
                in_brace = True
        elif part.endswith("}") and in_brace:
            yield part[:-1]
            in_brace = False
        else:
            yield part


@dataclass
class Image(Element):
    """An image in Markdown.

    Represents an image with alt text, URL, and attributes.

    Example:
        `![Alt text](image.png){#id .class width=100}`
    """

    pattern = re.compile(
        r"(?<![`])!\[(?P<alt>.*?)\]\((?P<url>.*?)\)\{(?P<attr>.*?)\}(?![`])",
        re.MULTILINE | re.DOTALL,
    )

    alt: str = ""

    @classmethod
    def from_match(cls, match: re.Match[str]) -> Self:
        identifier, classes, attributes = _parse(match.group("attr"))

        source = ""

        for k, cls_ in enumerate(classes):
            if cls_.startswith("`") and cls_.endswith("`"):
                source = cls_[1:-1]
                classes = classes[:k] + classes[k + 1 :]
                break

        return cls(
            match.group(0),
            identifier,
            classes,
            attributes,
            source=source,
            url=match.group("url"),
            alt=match.group("alt"),
        )


def parse(
    text: str,
    pos: int = 0,
    endpos: int | None = None,
    classes: tuple[type[Matcher], ...] = (Comment, CodeBlock, Image),
) -> Iterator[CodeBlock | Image | str]:
    """Parse the text and yield elements.

    Finds all occurrences of the specified element types in the text and
    yields either element instances or text segments between elements.

    Args:
        text (str): The text to search.
        pos (int): The starting position.
        endpos (int | None): The ending position.
        classes (tuple[type[Element], ...]): The element types to find.

    Yields:
        Element | str: Element instances or text segments.
    """
    if not classes:
        yield text[pos:endpos]
        return

    indent = ""
    for elem in classes[0].iter_elements(text, pos, endpos):
        if isinstance(elem, str):
            yield elem

        elif isinstance(elem, tuple):
            if indent:
                raise NotImplementedError

            indent = _get_indent(text[elem[0] : elem[1]])
            yield from parse(text, elem[0], elem[1] - len(indent), classes[1:])

        elif isinstance(elem, CodeBlock | Image):
            if isinstance(elem, Image):
                elem.indent = indent
                elem.text = f"{indent}{elem.text}"
                indent = ""
            elif indent:
                raise NotImplementedError
            yield elem

        else:
            raise NotImplementedError


def _get_indent(text: str) -> str:
    if "\n" not in text:
        return ""

    indent = text.rsplit("\n", 1)[-1]
    if all(c.isspace() for c in indent):
        return indent

    return ""


@cache
def get_language(text: str) -> str:
    """Get the language of a Markdown document.

    Determines the primary programming language used in the document's
    code blocks.

    Args:
        text (str): The Markdown text.

    Returns:
        str: The detected language, or "python" if not detected.
    """
    languages = {}
    identifiers = []

    for elem in parse(text):
        if isinstance(elem, CodeBlock) and elem.identifier and elem.classes:
            language = elem.classes[0].removeprefix(".")
            languages[elem.identifier] = language
        elif isinstance(elem, Image) and elem.identifier and elem.url in (".md", ""):
            if elem.identifier in languages:
                return languages[elem.identifier]
            identifiers.append(elem.identifier)

    for identifier in identifiers:
        if identifier in languages:
            return languages[identifier]

    if languages:
        return next(iter(languages.values()))

    return ""


def is_target_code_block(elem: Element | str, language: str) -> TypeGuard[CodeBlock]:
    """Check if an element is a code block with the target language.

    Args:
        elem (Element | str): The element to check.
        language (str): The target language.

    Returns:
        TypeGuard[CodeBlock]: True if the element is a code block with the
            target language, False otherwise.
    """
    if not language:
        return False

    if not isinstance(elem, CodeBlock):
        return False

    if not elem.identifier:
        return False

    return bool(elem.classes and elem.classes[0] in (language, f".{language}"))


def new_notebook(text: str) -> NotebookNode:
    """Create a new notebook from Markdown text.

    Parses the Markdown text, extracts code blocks, and creates a notebook
    with a code cell for each code block.

    Args:
        text (str): The Markdown text.

    Returns:
        NotebookNode: The created notebook.
    """
    language = get_language(text)

    if not language:
        msg = "language not found"
        raise ValueError(msg)

    node = nbformat.v4.new_notebook()
    node["metadata"]["language_info"] = {"name": language}

    for code_block in parse(text):
        if is_target_code_block(code_block, language):
            source = f"# #{code_block.identifier}\n{code_block.source}"
            cell = nbformat.v4.new_code_cell(source)
            node["cells"].append(cell)

    return node



"""Formatters for visualization outputs in notebooks.

This module provides functions for converting visualization outputs from
various libraries (matplotlib, seaborn, holoviews) to different formats
(PGF, PDF, SVG), and utilities for registering these formatters with IPython.
"""

from __future__ import annotations

import base64
import io
import tempfile
import uuid
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from IPython.lib.pretty import RepresentationPrinter
    from matplotlib.figure import Figure
    from seaborn.objects import Plot


def matplotlib_figure_to_pgf(fig: Figure, rp: RepresentationPrinter, cycle) -> None:
    """Convert a matplotlib figure to PGF format.

    Creates a temporary PGF file, reads its content, encodes embedded images
    as base64, and outputs the result using the representation printer.

    Args:
        fig (Figure): The matplotlib figure to convert.
        rp (RepresentationPrinter): The printer to output the result.
        cycle: Object passed by IPython for object representation.
    """
    name = str(uuid.uuid4())

    with tempfile.TemporaryDirectory() as dirname:
        directory = Path(dirname)
        filename = f"{name}.pgf"
        path = directory / filename
        fig.savefig(path, format="pgf", bbox_inches="tight")
        text = path.read_text(encoding="utf-8")

        imagenames = [x.name for x in directory.iterdir() if x.name != filename]
        if not imagenames:
            return rp.text(text)

        for name in imagenames:
            text = text.replace(name, _encode_pgf_text(name, directory))

        return rp.text(text)


def _encode_pgf_text(name: str, directory: Path) -> str:
    """Encode an image file as a base64 data URL.

    Args:
        name (str): The name of the image file.
        directory (Path): The directory containing the image file.

    Returns:
        str: The base64-encoded data URL.
    """
    ext = name.rsplit(".", 1)[-1]
    data = (directory / name).read_bytes()
    b64 = base64.b64encode(data).decode()
    return f"data:image/{ext};base64,{b64}"


def matplotlib_figure_to_pdf(fig: Figure) -> bytes:
    """Convert a matplotlib figure to PDF format.

    Args:
        fig (Figure): The matplotlib figure to convert.

    Returns:
        bytes: The PDF content as bytes.
    """
    with io.BytesIO() as fp:
        fig.savefig(fp, format="pdf", bbox_inches="tight")
        return fp.getvalue()


def matplotlib_figure_to_svg(fig: Figure) -> str:
    """Convert a matplotlib figure to SVG format.

    Args:
        fig (Figure): The matplotlib figure to convert.

    Returns:
        str: The SVG content as a string.
    """
    with io.StringIO() as fp:
        fig.savefig(fp, format="svg", bbox_inches="tight")
        return fp.getvalue()


def seaborn_plot_to_pgf(plot: Plot, rp: RepresentationPrinter, cycle) -> None:
    """Convert a seaborn plot to PGF format.

    Args:
        plot (Plot): The seaborn plot to convert.
        rp (RepresentationPrinter): The printer to output the result.
        cycle: Object passed by IPython for object representation.
    """
    from seaborn._core.plot import theme_context

    plotter = plot.plot()
    with theme_context(plotter._theme):  # type: ignore
        return matplotlib_figure_to_pgf(plotter._figure, rp, cycle)  # type: ignore


def seaborn_plot_to_pdf(plot: Plot) -> bytes:
    """Convert a seaborn plot to PDF format.

    Args:
        plot (Plot): The seaborn plot to convert.

    Returns:
        bytes: The PDF content as bytes.
    """
    from seaborn._core.plot import theme_context

    plotter = plot.plot()
    with theme_context(plotter._theme):  # type: ignore
        return matplotlib_figure_to_pdf(plotter._figure)  # type: ignore


def seaborn_plot_to_svg(plot: Plot) -> str:
    """Convert a seaborn plot to SVG format.

    Args:
        plot (Plot): The seaborn plot to convert.

    Returns:
        str: The SVG content as a string.
    """
    from seaborn._core.plot import theme_context

    plotter = plot.plot()
    with theme_context(plotter._theme):  # type: ignore
        return matplotlib_figure_to_svg(plotter._figure)  # type: ignore


MIMES: dict[str, str] = {
    "pgf": "text/plain",
    "pdf": "application/pdf",
    "svg": "image/svg+xml",
}

MODULE_CLASSES: dict[str, list[tuple[str, str]]] = {
    "matplotlib": [("matplotlib.figure", "Figure")],
    "seaborn": [("seaborn._core.plot", "Plot")],
}

FUNCTIONS: dict[tuple[str, str], dict[str, Callable]] = {
    ("matplotlib.figure", "Figure"): {
        "pgf": matplotlib_figure_to_pgf,
        "pdf": matplotlib_figure_to_pdf,
        "svg": matplotlib_figure_to_svg,
    },
    ("seaborn._core.plot", "Plot"): {
        "pgf": seaborn_plot_to_pgf,
        "pdf": seaborn_plot_to_pdf,
        "svg": seaborn_plot_to_svg,
    },
}


try:
    from holoviews.ipython.display_hooks import display_hook, image_display

    @display_hook
    def pgf_display(element, max_frames):
        """Used to render elements to PGF if requested in the display formats.

        Args:
            element: The holoviews element to render.
            max_frames: Maximum number of frames to render.

        Returns:
            The rendered element.
        """
        return image_display(element, max_frames, fmt="pgf")

except ModuleNotFoundError:  # no cov
    pgf_display = None  # type: ignore


def set_formatter_holoviews(fmt: str) -> None:
    """Set a formatter for holoviews.

    Registers a formatter for holoviews elements in the specified format.

    Args:
        fmt (str): The format to use ('pgf', 'pdf', 'svg').
    """
    from holoviews.core.dimension import LabelledData
    from holoviews.core.options import Store

    if fmt not in Store.display_formats:
        Store.display_formats.append(fmt)

    if fmt == "pgf" and pgf_display:
        Store.set_display_hook(fmt, LabelledData, pgf_display)


def set_formatter(module: str, fmt: str, ip=None) -> None:
    """Set a formatter for visualization outputs.

    Registers a formatter for the specified module and format with IPython.

    Args:
        module (str): The module to set the formatter for ('matplotlib',
            'seaborn', 'holoviews').
        fmt (str): The format to use ('pgf', 'pdf', 'svg').
        ip: Optional IPython instance. If None, get_ipython() is used.

    Raises:
        NotImplementedError: If the format is not supported.
        ModuleNotFoundError: If IPython is not installed.
    """
    if module == "holoviews":
        set_formatter_holoviews(fmt)
        return

    try:
        from IPython.core.getipython import get_ipython
    except ModuleNotFoundError:  # no cov
        msg = "IPython is not installed"
        raise ModuleNotFoundError(msg) from None

    if not ip and not (ip := get_ipython()):
        return

    if not (mime := MIMES.get(fmt)):
        raise NotImplementedError

    formatter = ip.display_formatter.formatters[mime]  # type:ignore

    if module_classes := MODULE_CLASSES.get(module):
        for module_class in module_classes:
            if function := FUNCTIONS.get(module_class, {}).get(fmt):
                formatter.for_type_by_name(*module_class, function)



"""Core functionality for reading and managing notebook files.

This module provides the main interface for accessing notebook content
from various file formats and managing notebook instances in memory.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import nbformat

import nbstore.markdown
import nbstore.python

if TYPE_CHECKING:
    from collections.abc import Iterable

    from nbformat import NotebookNode


class Store:
    """Manage notebook files from one or more source directories.

    Provides a centralized interface for reading notebook files and caching
    their content for efficient access. Automatically reloads files when
    they have been modified on disk.

    Attributes:
        src_dirs: List of source directories to search for notebook files.
        nodes: Dictionary mapping file paths to their notebook nodes.
        st_mtime: Dictionary mapping file paths to their last modification times.
        url: String representing the last accessed URL.
    """

    src_dirs: list[Path]
    nodes: dict[Path, NotebookNode]
    st_mtime: dict[Path, float]
    url: str

    def __init__(self, src_dirs: str | Path | Iterable[str | Path]) -> None:
        """Initialize a new Store instance.

        Args:
            src_dirs (str | Path | Iterable[str | Path]): One or more directories
                to search for notebook files. Can be a single path or a collection
                of paths.
        """
        if isinstance(src_dirs, (str, Path)):
            src_dirs = [src_dirs]

        self.src_dirs = [Path(src_dir) for src_dir in src_dirs]
        self.nodes = {}
        self.st_mtime = {}
        self.url = ""

    def find_path(self, url: str) -> Path:
        """Find the absolute path of a notebook file.

        Searches for the notebook file in the source directories. If the URL is
        an absolute path, it is returned directly.

        Args:
            url (str): The URL or relative path of the notebook file.

        Returns:
            Path: The absolute path to the notebook file.

        Raises:
            ValueError: If the file cannot be found in any source directory.
        """
        if Path(url).is_absolute():
            return Path(url)

        for src_dir in self.src_dirs:
            abs_path = (src_dir / url).absolute()
            if abs_path.exists():
                return abs_path

        msg = f"Source file not found in any source directory: {url}"
        raise ValueError(msg)

    def read(self, url: str) -> NotebookNode:
        """Read a notebook file and return its content.

        If the file has been modified since it was last read, it is reloaded.
        If no URL is provided, the last URL is used.

        Args:
            url (str): The URL or relative path of the notebook file.

        Returns:
            NotebookNode: The notebook content.
        """
        url = self.url = url or self.url

        path = self.find_path(url)
        st_mtime = path.stat().st_mtime

        if self.st_mtime.get(path) != st_mtime:
            self.nodes[path] = read(path)
            self.st_mtime[path] = st_mtime

        return self.nodes[path]

    def write(self, url: str, notebook_node: NotebookNode) -> None:
        """Write a notebook node to a file.

        Currently only supports writing to .ipynb files.

        Args:
            url (str): The URL or relative path of the notebook file.
            notebook_node (NotebookNode): The notebook content to write.

        Raises:
            NotImplementedError: If the file format is not supported for writing.
        """
        url = url or self.url

        path = self.find_path(url)

        if path.suffix == ".ipynb":
            return nbformat.write(notebook_node, path)

        raise NotImplementedError


def read(path: str | Path) -> NotebookNode:
    """Read a notebook file and return its content.

    Supports .ipynb, .py, and .md file formats.

    Args:
        path (str | Path): The path to the notebook file.

    Returns:
        NotebookNode: The notebook content.

    Raises:
        NotImplementedError: If the file format is not supported.
    """
    path = Path(path)

    if path.suffix == ".ipynb":
        return nbformat.read(path, as_version=4)  # type: ignore

    text = path.read_text()

    if path.suffix == ".py":
        return nbstore.python.new_notebook(text)

    if path.suffix == ".md":
        return nbstore.markdown.new_notebook(text)

    raise NotImplementedError



from pathlib import Path

import nbformat
import pytest
from nbformat import NotebookNode


@pytest.fixture(scope="module")
def nb():
    path = Path(__file__).parent.joinpath("mime.ipynb")
    nb = nbformat.read(path, as_version=4)
    assert isinstance(nb, NotebookNode)
    return nb


def test_mime(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "plot")
    assert len(data) == 3
    assert "text/plain" in data
    assert "image/svg+xml" in data
    assert "image/png" in data


def test_text(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "text")
    assert len(data) == 3
    assert "text/plain" in data
    assert "image/svg+xml" in data
    assert "image/png" in data
    assert "font-family: 'DejaVu Sans', 'Harano Aji Gothic';" in data["image/svg+xml"]



import io
import re
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from IPython.lib.pretty import RepresentationPrinter
from PIL import Image


@pytest.fixture(scope="module")
def text():
    from nbstore.formatter import matplotlib_figure_to_pgf

    fig, axes = plt.subplots(1, 2, figsize=(4, 2))
    for i in [0, 1]:
        data = np.random.randn(10, 10)
        axes[i].imshow(data, interpolation="nearest", aspect=1)
        axes[i].set(xlabel="x", ylabel="α")

    out = io.StringIO()
    rp = RepresentationPrinter(out)

    matplotlib_figure_to_pgf(fig, rp, None)
    return out.getvalue()


def test_matplotlib_figure_to_pgf_raster(text: str):
    assert text.count("]{data:image/png;base64,iVBOR") == 2


def test_findall(text: str):
    from nbstore.notebook import BASE64_PATTERN

    assert len(BASE64_PATTERN.findall(text)) == 2


def test_convert(text: str):
    from nbstore.notebook import _convert_pgf

    text = _convert_pgf(text)

    k = 0

    for filename in re.findall(r"\{\\includegraphics\[.+?\]\{(.+?)\}\}", text):
        assert isinstance(filename, str)
        assert filename.endswith(".png")
        assert Path(filename).exists()
        image = Image.open(filename)
        assert image.format == "PNG"
        assert image.size == (141, 141)
        k += 1

    assert k == 2


def test_convert_none():
    from nbstore.notebook import _convert_pgf

    assert _convert_pgf("abc") == "abc"



import nbformat
import pytest
from nbformat import NotebookNode

SOURCE = """\
# #html
from IPython.display import HTML
HTML("<p><strong>Hello, World!</strong></p>")
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [nbformat.v4.new_code_cell(SOURCE)]
    execute(nb)
    return nb


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "html")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/html" in data
    assert data["text/html"] == "<p><strong>Hello, World!</strong></p>"


def test_mime_content(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "html")
    assert isinstance(content, tuple)
    assert len(content) == 2
    assert isinstance(content[1], str)
    assert content[0] == "text/html"
    assert content[1] == "<p><strong>Hello, World!</strong></p>"



from pathlib import Path

import nbformat
import pytest

from nbstore.store import Store

SOURCE = """\
# #fig
import matplotlib.pyplot as plt
plt.plot([1, 10, 100], [-1, 0, 1])
"""


@pytest.fixture(scope="module")
def store(tmp_path_factory: pytest.TempPathFactory):
    src_dir = tmp_path_factory.mktemp("test")
    nb = nbformat.v4.new_notebook()
    nb["cells"] = [nbformat.v4.new_code_cell(SOURCE)]
    nbformat.write(nb, src_dir / "a.ipynb")
    return Store(src_dir)


def test_find_path(store: Store):
    path = store.find_path("a.ipynb")
    assert path.name == "a.ipynb"
    path = store.find_path(path.absolute().as_posix())
    assert path.name == "a.ipynb"


def test_find_path_error(store: Store):
    with pytest.raises(ValueError, match="Source file not found"):
        store.find_path("unknown")


def test_read(store: Store):
    from nbstore.notebook import get_source

    nb = store.read("a.ipynb")
    assert get_source(nb, "fig").startswith("import matplotlib.pyplot as plt")
    nb = store.read("")
    assert get_source(nb, "fig").startswith("import matplotlib.pyplot as plt")
    assert store.url == "a.ipynb"


def test_write(store: Store):
    from nbstore.notebook import get_source

    nb = store.read("a.ipynb")
    nb["cells"].append(nbformat.v4.new_code_cell("# #test\n123"))
    store.write("a.ipynb", nb)
    nb = store.read("a.ipynb")
    assert get_source(nb, "test") == "123"


def test_read_python(tmp_path: Path):
    from nbstore.notebook import get_source
    from nbstore.store import read

    path = tmp_path / "test.py"
    path.write_text("# %% #id1\nprint(1)\n# %% #id2\nprint(2)")
    nb = read(path)
    assert get_source(nb, "id1") == "print(1)"
    assert get_source(nb, "id2") == "print(2)"


def test_read_markdown(tmp_path: Path):
    from nbstore.notebook import get_language, get_source
    from nbstore.store import read

    path = tmp_path / "test.md"
    path.write_text("```julia #id1\nprintln(1)\n```\n")
    nb = read(path)
    assert get_language(nb) == "julia"
    assert get_source(nb, "id1") == "println(1)"



import nbformat
import pytest
from nbformat import NotebookNode

SOURCE_FORMATTER = """\
import matplotlib_inline.backend_inline
from nbstore.formatter import set_formatter
matplotlib_inline.backend_inline.set_matplotlib_formats("retina")
set_formatter("matplotlib", "svg")
"""

SOURCE_FIG = """\
# #fig
import matplotlib.pyplot as plt
import numpy as np
data = np.random.randn(50, 50)
fig, axes = plt.subplots(1, 2, figsize=(3, 2), gridspec_kw={"width_ratios": [1, 2]})
axes[0].imshow(data, interpolation="nearest", aspect=1)
axes[1].imshow(data, interpolation="nearest", aspect=1)
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [
        nbformat.v4.new_code_cell(SOURCE_FORMATTER),
        nbformat.v4.new_code_cell(SOURCE_FIG),
    ]
    execute(nb)
    return nb


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "fig")
    assert isinstance(data, dict)
    assert len(data) == 3
    assert "image/svg+xml" in data


@pytest.fixture(scope="module")
def svg(nb: NotebookNode):
    from nbstore.notebook import get_data

    return get_data(nb, "fig")["image/svg+xml"]


def test_svg(svg: str):
    assert svg.startswith('<?xml version="1.0" encoding="utf-8" standalone="no"?>')
    assert svg.count('<image xlink:href="data:image/png;base64,') == 2



import nbformat
import pytest
from nbformat import NotebookNode

SOURCE_FORMATTER = """\
import matplotlib_inline.backend_inline
from nbstore.formatter import set_formatter
matplotlib_inline.backend_inline.set_matplotlib_formats("retina")
set_formatter("matplotlib", "pgf")
"""

SOURCE_FIG = """\
# #fig
import matplotlib.pyplot as plt
plt.plot([1, 10, 100], [-1, 0, 1])
"""

SOURCE_STREAM = """\
# #stream
print(123)
"""

SOURCE_EMPTY = """\
# #empty
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [
        nbformat.v4.new_code_cell(SOURCE_FORMATTER),
        nbformat.v4.new_code_cell(SOURCE_FIG),
        nbformat.v4.new_code_cell(SOURCE_STREAM),
        nbformat.v4.new_code_cell(SOURCE_EMPTY),
    ]
    execute(nb)
    return nb


def test_cell(nb: NotebookNode):
    from nbstore.notebook import get_cell

    cell = get_cell(nb, "fig")
    assert isinstance(cell, dict)
    assert "cell_type" in cell


def test_cell_error(nb: NotebookNode):
    from nbstore.notebook import get_cell

    with pytest.raises(ValueError, match="Unknown identifier: unknown"):
        get_cell(nb, "unknown")


def test_source(nb: NotebookNode):
    from nbstore.notebook import get_source

    source = get_source(nb, "fig")
    assert isinstance(source, str)
    assert source.startswith("import")
    assert "plt.plot" in source


def test_source_with_identifier(nb: NotebookNode):
    from nbstore.notebook import get_source

    source = get_source(nb, "fig", include_identifier=True)
    assert isinstance(source, str)
    assert source.startswith("# #fig\n")


def test_outputs(nb: NotebookNode):
    from nbstore.notebook import get_outputs

    outputs = get_outputs(nb, "fig")
    assert isinstance(outputs, list)
    assert len(outputs) == 2


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "fig")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data
    assert "image/png" in data
    assert data["text/plain"].startswith("%% Creator: Matplotlib,")


def test_stream(nb: NotebookNode):
    from nbstore.notebook import get_stream

    stream = get_stream(nb, "stream")
    assert isinstance(stream, str)
    assert stream == "123\n"


def test_stream_none(nb: NotebookNode):
    from nbstore.notebook import get_stream

    assert get_stream(nb, "empty") is None


def test_data_empty(nb: NotebookNode):
    from nbstore.notebook import get_data

    assert get_data(nb, "empty") == {}


def test_mime_content_empty(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    assert get_mime_content(nb, "empty") == ("", "")



import re
from pathlib import Path

import nbformat
import pytest
from nbformat import NotebookNode
from PIL import Image

SOURCE_FORMATTER = """\
import matplotlib_inline.backend_inline
from nbstore.formatter import set_formatter
matplotlib_inline.backend_inline.set_matplotlib_formats("retina")
set_formatter("matplotlib", "pgf")
"""

SOURCE_FIG = """\
# #fig
import matplotlib.pyplot as plt
import numpy as np
data = np.random.randn(50, 50)
fig, axes = plt.subplots(1, 2, figsize=(3, 2), gridspec_kw={"width_ratios": [1, 2]})
axes[0].imshow(data, interpolation="nearest", aspect=1)
axes[1].imshow(data, interpolation="nearest", aspect=1)
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [
        nbformat.v4.new_code_cell(SOURCE_FORMATTER),
        nbformat.v4.new_code_cell(SOURCE_FIG),
    ]
    execute(nb)
    return nb


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "fig")
    assert isinstance(data, dict)
    assert len(data) == 2
    assert "text/plain" in data


@pytest.fixture(scope="module")
def text(nb: NotebookNode):
    from nbstore.notebook import get_data

    return get_data(nb, "fig")["text/plain"]


def test_backend(text: str):
    assert text.startswith("%% Creator: Matplotlib, PGF backend")


def test_convert(text: str):
    i = 0

    for k, filename in enumerate(
        re.findall(r"\{\\includegraphics\[.+?\]\{(.+?)\}\}", text),
    ):
        assert isinstance(filename, str)
        assert filename.endswith(".png")
        assert Path(filename).exists()
        image = Image.open(filename)
        assert image.format == "PNG"
        if k == 0:
            assert image.size == (71, 71)
        else:
            assert image.size == (141, 141)
        i += 1

    assert i == 2



import base64

import holoviews as hv
import polars as pl
import pytest
from holoviews.core.options import Store
from holoviews.plotting.mpl.renderer import MPLRenderer

hv.extension("matplotlib")  # type: ignore


@pytest.fixture(scope="module")
def curve():
    df = pl.DataFrame({"x": [1, 2], "y": [3, 4]})
    return hv.Curve(df, "x", "y")


@pytest.fixture(scope="module")
def renderer():
    return Store.renderers["matplotlib"]


@pytest.fixture(scope="module")
def plot(renderer: MPLRenderer, curve):
    return renderer.get_plot(curve)


def test_pgf(renderer: MPLRenderer, plot):
    from nbstore.formatter import set_formatter

    set_formatter("holoviews", "pgf")

    data, metadata = renderer(plot, fmt="pgf")
    assert isinstance(data, bytes)
    assert data.startswith(b"%% Creator: Matplotlib, PGF backend")
    assert isinstance(metadata, dict)
    assert metadata["mime_type"] == "text/pgf"


@pytest.mark.parametrize(
    ("fmt", "text", "mime"),
    [("png", "iVBOR", "image/png"), ("pdf", "JVBER", "application/pdf")],
)
def test_png(renderer: MPLRenderer, plot, fmt, text, mime):
    data, metadata = renderer(plot, fmt=fmt)
    assert isinstance(data, bytes)
    assert base64.b64encode(data).decode().startswith(text)
    assert metadata["mime_type"] == mime


def test_svg(renderer: MPLRenderer, plot):
    xml, metadata = renderer(plot, fmt="svg")
    assert isinstance(xml, str)
    assert xml.startswith('<?xml version="1.0"')
    assert metadata["mime_type"] == "image/svg+xml"



"""Utilities for working with Jupyter notebook content.

This module provides functions for extracting and manipulating content
from notebook cells, with a focus on cell identification by unique
identifiers and data extraction from cell outputs.
"""

from __future__ import annotations

import atexit
import base64
import re
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any

import nbformat

if TYPE_CHECKING:
    from nbformat import NotebookNode


def get_language(nb: NotebookNode, default: str = "python") -> str:
    """Get the programming language of a notebook.

    Extracts the language from notebook metadata, either from the kernelspec
    or language_info sections.

    Args:
        nb (NotebookNode): The notebook to examine.
        default (str): Default language to return if not found in metadata.

    Returns:
        str: The programming language name.
    """
    metadata = nb["metadata"]

    if "kernelspec" in metadata:
        return metadata["kernelspec"].get("language", default)

    if "language_info" in metadata:
        return metadata["language_info"].get("name", default)

    return default


def get_cell(nb: NotebookNode, identifier: str) -> dict[str, Any]:
    """Get a cell by its identifier.

    Searches for a cell whose source code starts with a specific identifier
    pattern, supporting both "# #" and "# %% #" prefixes.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier to look for.

    Returns:
        dict[str, Any]: The cell dictionary.

    Raises:
        ValueError: If no cell with the given identifier is found.
    """
    for cell in nb["cells"]:
        source: str = cell["source"]
        for prefix in ["# #", "# %% #", "#| label: ", "# | label: "]:
            if source.startswith(f"{prefix}{identifier}\n"):
                return cell

    msg = f"Unknown identifier: {identifier}"
    raise ValueError(msg)


def get_source(
    nb: NotebookNode,
    identifier: str,
    *,
    include_identifier: bool = False,
) -> str:
    """Get the source code of a cell by its identifier.

    Extracts the source code from a cell, optionally including the
    identifier line.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier of the cell.
        include_identifier (bool): Whether to include the identifier line.

    Returns:
        str: The source code of the cell.

    Raises:
        ValueError: If no cell with the given identifier is found.
        NotImplementedError: If the cell has no source.
    """
    if source := get_cell(nb, identifier).get("source", ""):
        if include_identifier:
            return source

        return source.split("\n", 1)[1]

    raise NotImplementedError


def get_outputs(nb: NotebookNode, identifier: str) -> list:
    """Get the outputs of a cell by its identifier.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier of the cell.

    Returns:
        list: The outputs of the cell.
    """
    return get_cell(nb, identifier).get("outputs", [])


def _get_data_by_type(outputs: list, output_type: str) -> dict[str, str] | None:
    """Get data from outputs by output type.

    Internal helper function to extract data from outputs based on the output type.

    Args:
        outputs (list): List of output dictionaries.
        output_type (str): The type of output to look for.

    Returns:
        dict[str, str] | None: The data dictionary, or None if not found.
    """
    for output in outputs:
        if output["output_type"] == output_type:
            if output_type == "stream":
                return {"text/plain": output["text"]}

            return output["data"]

    return None


def get_stream(nb: NotebookNode, identifier: str) -> str | None:
    """Get the stream output of a cell by its identifier.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier of the cell.

    Returns:
        str | None: The stream output text, or None if not found.
    """
    outputs = get_outputs(nb, identifier)
    if data := _get_data_by_type(outputs, "stream"):
        return data["text/plain"]

    return None


def get_data(nb: NotebookNode, identifier: str) -> dict[str, str]:
    """Get the data outputs of a cell by its identifier.

    Tries to find data in various output types, prioritizing display_data,
    execute_result, and stream in that order.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier of the cell.

    Returns:
        dict[str, str]: The data dictionary, empty if no data is found.
    """
    outputs = get_outputs(nb, identifier)

    for type_ in ["display_data", "execute_result", "stream"]:
        if data := _get_data_by_type(outputs, type_):
            return _convert_data(data)

    return {}


def _convert_data(data: dict[str, str]) -> dict[str, str]:
    """Convert special data formats.

    Currently handles conversion of Matplotlib PGF backend outputs.

    Args:
        data (dict[str, str]): The data dictionary to convert.

    Returns:
        dict[str, str]: The converted data dictionary.
    """
    text = data.get("text/plain")
    if text and text.startswith("%% Creator: Matplotlib, PGF backend"):
        data["text/plain"] = _convert_pgf(text)

    return data


BASE64_PATTERN = re.compile(r"\{data:image/(?P<ext>.*?);base64,(?P<b64>.*?)\}")


def _convert_pgf(text: str) -> str:
    """Convert embedded base64 images in PGF text to file references.

    Args:
        text (str): The PGF text with embedded base64 images.

    Returns:
        str: The converted PGF text with file references.
    """

    def replace(match: re.Match) -> str:
        ext = match.group("ext")
        data = base64.b64decode(match.group("b64"))

        with tempfile.NamedTemporaryFile(suffix=f".{ext}", delete=False) as tmp:
            tmp.write(data)
            path = Path(tmp.name)

        atexit.register(lambda p=path: p.unlink(missing_ok=True))

        return f"{{{path.absolute()}}}"

    return BASE64_PATTERN.sub(replace, text)


def get_mime_content(
    nb: NotebookNode,
    identifier: str,
) -> tuple[str, str | bytes]:
    """Get the MIME content of a cell by its identifier.

    Extracts the content of a cell output based on MIME type, prioritizing
    SVG, HTML, PDF, other images, and plain text in that order.

    Args:
        nb (NotebookNode): The notebook to search.
        identifier (str): The identifier of the cell.

    Returns:
        tuple[str, str | bytes]: A tuple of (mime_type, content),
            or ("", "") if no content is found.
    """
    data = get_data(nb, identifier)
    for mime in ["image/svg+xml", "text/html"]:
        if text := data.get(mime):
            return mime, text

    if text := data.get("application/pdf"):
        return "application/pdf", base64.b64decode(text)

    for mime, text in data.items():
        if mime.startswith("image/"):
            return mime, base64.b64decode(text)

    if "text/plain" in data:
        return "text/plain", data["text/plain"]

    return "", ""


def add_data(nb: NotebookNode, identifier: str, mime: str, data: str) -> None:
    """Add data to a cell output by its identifier.

    Adds or updates data in a display_data output of a cell.

    Args:
        nb (NotebookNode): The notebook to modify.
        identifier (str): The identifier of the cell.
        mime (str): The MIME type of the data.
        data (str): The data content.
    """
    outputs = get_outputs(nb, identifier)
    if output := _get_data_by_type(outputs, "display_data"):
        output[mime] = data


def new_code_cell(identifier: str, source: str) -> NotebookNode:
    """Create a new code cell with an identifier.

    If the source doesn't already have an identifier, one is added.

    Args:
        identifier (str): The identifier for the cell.
        source (str): The source code.

    Returns:
        NotebookNode: A new code cell.
    """
    if not source.startswith("#") or f"#{identifier}" not in source.split("\n", 1)[0]:
        source = f"# #{identifier}\n{source}"

    return nbformat.v4.new_code_cell(source)


def execute(
    nb: NotebookNode,
    timeout: int = 600,
) -> tuple[NotebookNode, dict[str, Any]]:
    """Execute a notebook.

    Uses nbconvert's ExecutePreprocessor to run all cells in a notebook.

    Args:
        nb (NotebookNode): The notebook to execute.
        timeout (int): Maximum time in seconds to wait for each cell execution.

    Returns:
        tuple[NotebookNode, dict[str, Any]]: The executed notebook and execution info.

    Raises:
        ModuleNotFoundError: If nbconvert is not installed.
    """
    try:
        from nbconvert.preprocessors.execute import ExecutePreprocessor
    except ModuleNotFoundError:  # no cov
        msg = "nbconvert is not installed"
        raise ModuleNotFoundError(msg) from None

    ep = ExecutePreprocessor(timeout=timeout)
    return ep.preprocess(nb)


def equals(nb: NotebookNode, other: NotebookNode) -> bool:
    """Check if two notebooks have the same cells.

    Compares the number of cells and the source of each cell.

    Args:
        nb (NotebookNode): First notebook.
        other (NotebookNode): Second notebook.

    Returns:
        bool: True if the notebooks have the same cells, False otherwise.
    """
    if len(nb["cells"]) != len(other["cells"]):
        return False

    for cell1, cell2 in zip(nb["cells"], other["cells"], strict=False):
        if cell1["source"] != cell2["source"]:
            return False

    return True



"""Convert Python scripts to Jupyter notebooks.

This module provides utilities for parsing Python scripts and converting them
to notebook format, with support for cell markers and code block extraction.
"""

from __future__ import annotations

import re
import textwrap
from typing import TYPE_CHECKING

import nbformat

if TYPE_CHECKING:
    from collections.abc import Iterator

    from nbformat import NotebookNode


def _split_indent(text: str) -> Iterator[str]:
    """Split text at the first unindented line after indented lines.

    Handles code blocks with mixed indentation by splitting at the first
    non-indented line after indented lines. This is useful for extracting
    code blocks like if __name__ == "__main__" sections.

    Args:
        text (str): The text to split.

    Yields:
        str: The split text blocks.
    """
    lines = text.split("\n")

    for line in lines:
        if not line.strip():
            continue
        if line.startswith((" ", "\t")):
            break
        else:
            yield text
            return

    for cursor, line in enumerate(lines):
        if not line.strip():
            continue
        if not line.startswith((" ", "\t")):
            block = "\n".join(lines[:cursor])
            yield textwrap.dedent(block)
            yield "\n".join(lines[cursor:])
            return

    yield textwrap.dedent(text)


def _iter(text: str, pattern: re.Pattern, *, dedent: bool = False) -> Iterator[str]:
    """Iterate through text blocks separated by pattern matches.

    Splits the text at lines matching the pattern and yields each block.

    Args:
        text (str): The text to iterate through.
        pattern (re.Pattern): The pattern to match.
        dedent (bool): Whether to dedent indented blocks.

    Yields:
        str: The text blocks.
    """
    start = 0
    lines = text.split("\n")

    for cursor, line in enumerate(lines):
        if pattern.match(line):
            if cursor > start:
                block = "\n".join(lines[start:cursor])
                if dedent:
                    yield from _split_indent(block)
                else:
                    yield block
            start = cursor + (1 if dedent else 0)

    if start < len(lines):
        block = "\n".join(lines[start:])
        if dedent:
            yield from _split_indent(block)
        else:
            yield block


MAIN_PATTERN = re.compile(r"if\s+__name__\s*==\s*['\"]__main__['\"]\s*:")


def _iter_main_blocks(text: str) -> Iterator[str]:
    """Iterate through main blocks in the text.

    Splits the text at 'if __name__ == "__main__":' lines and yields each block.

    Args:
        text (str): The text to iterate through.

    Yields:
        str: The main blocks.
    """
    yield from _iter(text, MAIN_PATTERN, dedent=True)


CELL_PATTERN = re.compile(r"# %%")


def _iter_sources(text: str) -> Iterator[str]:
    """Iterate through cell sources in the text.

    Splits the text at '# %%' cell markers and yields each cell.

    Args:
        text (str): The text to iterate through.

    Yields:
        str: The cell sources.
    """
    yield from _iter(text, CELL_PATTERN, dedent=False)


def parse(text: str) -> Iterator[str]:
    """Parse the text and yield sources.

    First splits the text at 'if __name__ == "__main__":' lines,
    then splits each block at '# %%' cell markers.

    Args:
        text (str): The text to parse.

    Yields:
        str: The sources.
    """
    for block in _iter_main_blocks(text):
        for source in _iter_sources(block):
            yield source.rstrip()


def new_notebook(text: str) -> NotebookNode:
    """Create a new notebook from Python code.

    Parses the Python code, extracts cell sources, and creates a notebook
    with a code cell for each source.

    Args:
        text (str): The Python code to convert.

    Returns:
        NotebookNode: The created notebook.
    """
    node = nbformat.v4.new_notebook()
    node["metadata"]["language_info"] = {"name": "python"}

    for source in parse(text):
        cell = nbformat.v4.new_code_cell(source)
        node["cells"].append(cell)

    return node



import nbformat
import pytest
from nbformat import NotebookNode

SOURCE_STDOUT = """\
# #stdout
a = "stdout"
a
"""

SOURCE_STREAM = """\
# #stream
print("stream1")
print("stream2")
"""

SOURCE_BOTH = """\
# #both
print("print")
"hello"
"""

SOURCE_PANDAS = """\
# #pandas
import pandas as pd
pd.DataFrame({"a": [1, 2, 3]})
"""

SOURCE_POLARS = """\
# #polars
import polars as pl
pl.DataFrame({"a": [1, 2, 3]})
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [
        nbformat.v4.new_code_cell(SOURCE_STDOUT),
        nbformat.v4.new_code_cell(SOURCE_STREAM),
        nbformat.v4.new_code_cell(SOURCE_BOTH),
        nbformat.v4.new_code_cell(SOURCE_PANDAS),
        nbformat.v4.new_code_cell(SOURCE_POLARS),
    ]
    execute(nb)
    return nb


def test_mime_content_stdout(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "stdout")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "'stdout'"


def test_mime_content_stream(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "stream")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "stream1\nstream2\n"


def test_mime_content_both(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "both")
    assert isinstance(content, tuple)
    assert content[0] == "text/plain"
    assert content[1] == "'hello'"


def test_mime_content_pandas(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "pandas")
    assert isinstance(content, tuple)
    assert content[0] == "text/html"
    assert isinstance(content[1], str)
    assert content[1].startswith("<div>")


def test_mime_content_polars(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    content = get_mime_content(nb, "polars")
    assert isinstance(content, tuple)
    assert content[0] == "text/html"
    assert isinstance(content[1], str)
    assert content[1].startswith("<div>")



import nbformat
import pytest
from nbformat import NotebookNode

SOURCE_FORMATTER = """\
import matplotlib_inline.backend_inline
from nbstore.formatter import set_formatter
matplotlib_inline.backend_inline.set_matplotlib_formats("retina")
set_formatter("matplotlib", "svg")
"""

SOURCE_FIG = """\
# #fig
import matplotlib.pyplot as plt
plt.plot([1, 10, 100], [-1, 0, 1])
"""


@pytest.fixture(scope="module")
def nb():
    from nbstore.notebook import execute

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [
        nbformat.v4.new_code_cell(SOURCE_FORMATTER),
        nbformat.v4.new_code_cell(SOURCE_FIG),
    ]
    execute(nb)
    return nb


def test_data(nb: NotebookNode):
    from nbstore.notebook import get_data

    data = get_data(nb, "fig")
    assert isinstance(data, dict)
    assert len(data) == 3
    assert "text/plain" in data
    assert "image/png" in data
    assert data["image/svg+xml"].startswith('<?xml version="1.0"')


def test_mime_content(nb: NotebookNode):
    from nbstore.notebook import get_mime_content

    mime_content = get_mime_content(nb, "fig")
    assert isinstance(mime_content, tuple)
    mime, content = mime_content
    assert mime == "image/svg+xml"
    assert isinstance(content, str)
    assert content.startswith('<?xml version="1.0"')



from pathlib import Path

import nbformat
from nbformat import NotebookNode


def test_language_kernel():
    from nbstore.notebook import get_language

    path = Path(__file__).parent.joinpath("kernel.ipynb")
    nb = nbformat.read(path, as_version=4)
    assert isinstance(nb, NotebookNode)
    assert get_language(nb) == "python"


def test_language_info():
    from nbstore.notebook import execute, get_language

    nb = nbformat.v4.new_notebook()
    execute(nb)
    assert get_language(nb) == "python"


def test_language_default():
    from nbstore.notebook import get_language

    nb = nbformat.v4.new_notebook()
    assert isinstance(nb, NotebookNode)
    assert get_language(nb, "julia") == "julia"


SOURCE = """\
# #fig
import matplotlib.pyplot as plt
plt.plot([1, 10, 100], [-1, 0, 1])
"""


def test_add_data():
    from nbstore.notebook import add_data, execute, get_data

    nb = nbformat.v4.new_notebook()
    nb["cells"] = [nbformat.v4.new_code_cell(SOURCE)]
    execute(nb)
    add_data(nb, "fig", "image/pdf", "test")
    data = get_data(nb, "fig")
    assert data["image/pdf"] == "test"


def test_new_code_cell():
    from nbstore.notebook import new_code_cell

    cell = new_code_cell("fig", 'print("test")')
    assert cell["source"] == '# #fig\nprint("test")'


def test_new_code_cell_with_identifier():
    from nbstore.notebook import new_code_cell

    cell = new_code_cell("fig", '# #fig abc\nprint("test")')
    assert cell["source"] == '# #fig abc\nprint("test")'


def test_equals():
    from nbstore.notebook import equals

    nb1 = nbformat.v4.new_notebook()
    nb1["cells"] = [nbformat.v4.new_code_cell("a")]
    nb2 = nbformat.v4.new_notebook()
    assert not equals(nb1, nb2)
    nb2["cells"] = [nbformat.v4.new_code_cell("a")]
    assert equals(nb1, nb2)
    nb2["cells"] = [nbformat.v4.new_code_cell("b")]
    assert not equals(nb1, nb2)



import base64
import io

import pytest
import seaborn.objects as so
from IPython.core.formatters import PDFFormatter, PlainTextFormatter, SVGFormatter
from IPython.core.interactiveshell import InteractiveShell
from IPython.lib.pretty import RepresentationPrinter
from seaborn.objects import Plot

from nbstore.formatter import FUNCTIONS


def get_func(fmt: str):
    functions = FUNCTIONS.get(("seaborn._core.plot", "Plot"))
    assert functions
    function = functions.get(fmt)
    assert function
    return function


@pytest.fixture(scope="module")
def plot():
    return so.Plot()


def test_pgf(plot: Plot):
    out = io.StringIO()
    rp = RepresentationPrinter(out)
    function = get_func("pgf")
    function(plot, rp, None)
    text = out.getvalue()
    assert text.startswith("%% Creator: Matplotlib, PGF backend")
    assert text.endswith("\\endgroup%\n")


def test_pdf(plot: Plot):
    function = get_func("pdf")
    data = function(plot)
    assert isinstance(data, bytes)
    assert base64.b64encode(data).decode().startswith("JVBER")


def test_svg(plot: Plot):
    function = get_func("svg")
    xml = function(plot)
    assert isinstance(xml, str)
    assert xml.startswith('<?xml version="1.0"')


def test_set_formatter_pgf():
    from nbstore.formatter import seaborn_plot_to_pgf, set_formatter

    ip = InteractiveShell()
    set_formatter("seaborn", "pgf", ip)
    formatter = ip.display_formatter.formatters["text/plain"]  # type:ignore
    assert isinstance(formatter, PlainTextFormatter)
    func = formatter.lookup_by_type("seaborn._core.plot.Plot")
    assert func is seaborn_plot_to_pgf


def test_set_formatter_pdf():
    from nbstore.formatter import seaborn_plot_to_pdf, set_formatter

    ip = InteractiveShell()
    set_formatter("seaborn", "pdf", ip)
    formatter = ip.display_formatter.formatters["application/pdf"]  # type:ignore
    assert isinstance(formatter, PDFFormatter)
    func = formatter.lookup_by_type("seaborn._core.plot.Plot")
    assert func is seaborn_plot_to_pdf


def test_set_formatter_svg():
    from nbstore.formatter import seaborn_plot_to_svg, set_formatter

    ip = InteractiveShell()
    set_formatter("seaborn", "svg", ip)
    formatter = ip.display_formatter.formatters["image/svg+xml"]  # type:ignore
    assert isinstance(formatter, SVGFormatter)
    func = formatter.lookup_by_type("seaborn._core.plot.Plot")
    assert func is seaborn_plot_to_svg
